/*

		Sky Window procedure
                             
*/

#include "sunclock.h"
#include "project.h"

double skyLimitMag = 5.5;				// Limiting magnitude of sky map

int skyShowName = TRUE;					// Show star names in sky map ?
double skyNameMag = 2;					// Maximum magnitude to show names in sky
int skyShowBflam = TRUE;				// Show Bayer/Flamsteed numbers in sky ?
double skyBflamMag = 3.5;				// Maximum magnitude to show Bayer/Flamsteed in sky
int skyShowDeep = TRUE;					// Show deep sky objects in sky ?
double skyDeepMag = 4.5;				// Maximum magnitude to show deep sky objects in sky
int skyShowConstellations = TRUE;		// Show constellations
int skyShowConbounds = TRUE;			// Show constellation boundaries ?
int skyShowConnames = TRUE;				// Show constellation names ?
int skyAlignConnames = FALSE;			// Align constellation names ?
int skyShowCoords = TRUE;				// Show coordinate grid ?
int skyShowPlanets = TRUE;				// Show solar system objects ?
int skyShowTiming = TRUE;				// Display calculation time ?

int precessionCalculation = PrecAuto;	// Precession calculation mode
struct planet planet_info[11];			// Calculated planetary information
static mapwindow skywin;				// Sky windoe projection definition

#pragma pack(1)							// starMapData must be byte-packed
struct starMapData {
	unsigned short lon;
	short lat;
	unsigned short mag;
	unsigned char spectral;
};
#pragma pack()

static HGLOBAL starMapRes = NULL;       // Star map resource global block handle
static HRSRC starMapHandle = NULL;		// Star map resource handle
static HBITMAP smbitmap = NULL;			// Screen restore bitmap
static HPALETTE starPal = NULL;			// Star colour rendering palette
static BOOL firstTick = FALSE;			// No repaint before first timer tick
static BOOL nextTick = FALSE;			// Always repaint immediately after first timer tick
static BOOL bailedOut = FALSE;			// Set when termination signal received
static long calculationNumber = 0;		// Number of calculations performed

static double skyLham;					// Local hour angle offset at last sky map generation
static int Flip = 1;					// 1 if Northern hemisphere, -1 if Southern
static COLORREF skycolour = RGB(0, 0, 0); // Sky colour for legends

static RECT tickerRect;					// Rectangle for ticker time display
static long tickerLast;					// Last time ticker activated
static HDC hTickerDC;					// DC for ticker

HICON planetIcons[10] = {NULL};			// Solar system object icon handles
static HICON starIcons[9] = {NULL};		// Star stamper icon handles
static HICON deepIcons[10] = {NULL};	// Deep sky object icons
static int pirefc = 0;                  // Planet icons reference count

int starCatalogue = 0;					// Star catalogue selection
int starQuality = 2;					// Star quality/time tradeoff
int showStarColours = FALSE;			// Show star colours ?

static char *starCatResName[] = {		// Star catalogue resource names (don't translate these)
	"Yale"                              // 0: Yale bright star catalogue
};

static TIME_ZONE_INFORMATION tzInfo;	// Time zone information
char *tzName[2] = { "", "" };			// Time zone names

#define BrightestMag -1

/*  LOADPLANETICONS  --  Load the planet icons from the resource file.  */

void loadPlanetIcons(void)
{
	if (pirefc < 1) {
		int i;

		for (i = 0; i < ELEMENTS(planetIcons); i++) {
			planetIcons[i] = LoadIcon(hInst, MAKEINTRESOURCE(IDR_PLANETS + i));
#ifdef Mdiag
			if (planetIcons[i] == NULL) {
				char s[80];
				wsprintf(s, "planetIcons[%d] alloc failed.\r\n", i);
				OutputDebugString(s);
			}
#endif
		}
		for (i = 0; i < ELEMENTS(starIcons); i++) {
			starIcons[i] = LoadIcon(hInst, MAKEINTRESOURCE(IDR_STARS + i));
#ifdef Mdiag
			if (starIcons[i] == NULL) {
				char s[80];
				wsprintf(s, "starIcons[%d] alloc failed.\r\n", i);
				OutputDebugString(s);
			}
#endif
		}
		for (i = 0; i < ELEMENTS(deepIcons); i++) {
			deepIcons[i] = LoadIcon(hInst, MAKEINTRESOURCE(IDR_DEEP + i));
#ifdef Mdiag
			if (deepIcons[i] == NULL) {
				char s[80];
				wsprintf(s, "deepIcons[%d] alloc failed.\r\n", i);
				OutputDebugString(s);
			}
#endif
		}
	}
	pirefc++;
}
/*  UNLOADPLANETICONS  --  Release the planet icons when we're done with them.  */

void unloadPlanetIcons(void)
{
	pirefc--;
	if (pirefc < 1) {
		int i;

		for (i = 0; i < ELEMENTS(planetIcons); i++) {
			if (planetIcons[i] != NULL) {
				DestroyIcon(planetIcons[i]);
				planetIcons[i] = NULL;
			}
		}
		for (i = 0; i < ELEMENTS(starIcons); i++) {
			if (starIcons[i] != NULL) {
				DestroyIcon(starIcons[i]);
				starIcons[i] = NULL;
			}
		}
		for (i = 0; i < ELEMENTS(deepIcons); i++) {
			if (deepIcons[i] != NULL) {
				DestroyIcon(deepIcons[i]);
				deepIcons[i] = NULL;
			}
		}
	}
}

/*  CALCPLANET  --  Calculate planetary positions and altitude and azimuth from
					viewer's position.  */

void calcPlanet(double jd)
{
	int i;
	double igmst, latsin, latcos;

	planets(jd, 0xFFFF, planet_info);
	igmst = gmst(jd);
	latsin = sin(dtr(siteLat));
	latcos = cos(dtr(siteLat));
	for (i = 0; i <= 9; i++) {
		planet_info[i].lha = dtr(fixangle((igmst * 15) - siteLon - planet_info[i].ra));
		planet_info[i].az = rtd(atan2(sin(planet_info[i].lha), cos(planet_info[i].lha) * latsin -
								tan(dtr(planet_info[i].dec)) * latcos));
		planet_info[i].alt = rtd(asin(latsin * sin(dtr(planet_info[i].dec)) +
								latcos * cos(dtr(planet_info[i].dec)) * cos(planet_info[i].lha)));
	}
}

//  SET_TM_TIME  --  Set time from Windows system or local time

static void set_tm_time(struct tm *t, BOOL islocal)
{
#define CtF(tf, sf) t->tf = s.sf

	SYSTEMTIME s;

	if (islocal) {
		GetLocalTime(&s);
	} else {
		GetSystemTime(&s);
	}
	CtF(tm_sec, wSecond);
	CtF(tm_min, wMinute);
	CtF(tm_hour, wHour);
	CtF(tm_mday, wDay);
	CtF(tm_mon, wMonth - 1);
	CtF(tm_year, wYear - 1900);
	CtF(tm_wday, wDayOfWeek);
	//  tm_yday  never used
	t->tm_isdst = GetTimeZoneInformation(&tzInfo) == TIME_ZONE_ID_DAYLIGHT;
}

// TICKER  --  Periodically update seconds counter while calculating sky

static void ticker(void)
{
	SYSTEMTIME s;
	static WORD tickerLast;

	GetSystemTime(&s);
	if (s.wSecond != tickerLast && !fChildPreview) {
		struct tm ct, lt;
		
		tickerLast = s.wSecond;
		set_tm_time(&lt, TRUE);
		set_tm_time(&ct, FALSE);
		ssavetime(hTickerDC, lt, ct, tickerRect, skycolour);	// Update local and universal time
    }
}

/*	BAILOUT  --  See if we should bail out of calculation early.
				 Note that we don't actually process a pending
				 keyboard or mouse event, simply notify the caller
				 that one is queued.  This allows the screen saver
				 main program to actually receive the event and
				 perform its shut down. */

static BOOL bailout(void)
{
	MSG msg;
	
	if (bailedOut ||
		PeekMessage(&msg, hWndMain, WM_KEYFIRST, WM_KEYLAST, PM_NOREMOVE | PM_NOYIELD) ||
		PeekMessage(&msg, hWndMain, WM_MOUSEFIRST, WM_MOUSELAST, PM_NOREMOVE | PM_NOYIELD)) {
#ifdef Mdiag
		char s[120];
		
		if (!bailedOut) {
			wsprintf(s, "Bailed on %X, hWnd = %x (oWnd = %x) wParam = %X, lParam = %lX\r\n", msg.message, msg.hwnd, hMainWindow, msg.wParam, msg.lParam);
			OutputDebugString(s);
		} else {
			OutputDebugString("Redundant bailout.\r\n");	
		}
#endif		
		return bailedOut = TRUE;
	}
	return FALSE; 
}

/* PAINTSKY  --  Paint the sky window.  */

static HDC hhDC;					// Hidden HDC argument to D_movedraw

static void plotLine(double fdec, double fra, double tdec, double tra)
{
	int vx, vy, vx2, vy2;
	double cdec1, cra1, cdec2, cra2;

	if (clipr_xform(fdec, fra, tdec, tra, &vx, &vy, &vx2, &vy2, FALSE,
		&cdec1, &cra1, &cdec2, &cra2)) {
		drawcurveline(cdec1, cra1, cdec2, cra2, vx, vy, vx2, vy2, 0, FALSE, 0, hhDC);
	}
}

static HRSRC rscHand;
static HGLOBAL rscHandle = NULL;
static LPSTR rscPtr, rscPtrS;

static BOOL openRSC(LPSTR rname, LPSTR rtype)
{
	rscHand = FindResource(hInst, rname, rtype);
	if (rscHand != NULL) {
		rscHandle = LoadResource(hInst, rscHand);
		if (rscHandle != NULL) {
			rscPtr = LockResource(rscHandle);
			if (rscPtr != NULL) {
				rscPtrS = rscPtr;
				return TRUE;
			}
			FreeResource(rscHandle);		
		}
	} 
	return FALSE; 
}

static LPSTR fgetsRSC(LPSTR cl, int clen)
{
	int cn = 0;

	while (TRUE) {
		char c = *rscPtrS++;

		if (c != '\r') {
			if (cn < clen) {
				cl[cn++] = c;
			}
			if (c == '\n') {
				return cl;
			}	
		}
	}
}

static void closeRSC(void)
{
	if (rscHandle != NULL) {
		FreeResource(rscHandle);
		rscHandle = NULL;
	}
} 

static void FAR paintSky(int skyTel, HDC hDC, RECT *br, double limag)
{
	int i, vx, vy, vis, vx2, vy2, smhand = -1, pflip, precess;
	void far *smap;
	double igmst, lham, ra, dec, mag;
	unsigned char smex[4];

    hhDC = hDC;
	if ((precessionCalculation == PrecAlways) ||
		((precessionCalculation == PrecAuto) && (abs(PrecEpoch - faketime) >
			(PrecYears * 365.25)))) {
		definePrecession(2000.0 + ((faketime - PrecEpoch) / 365.25));
		precess = TRUE;
	} else {
		precess = FALSE;
	}
#define Prd(x, y) if (precess) { precessObject(x, y, &x, &y); }

	igmst = gmst(faketime);
    lham = skyLham = dtr((igmst * 15) - siteLon);
	skywin.width = br->right - br->left;
	skywin.height = br->bottom - br->top;
	skywin.x_offset = br->left;
	skywin.y_offset = br->top;
	skywin.proj_mode = STEREOGR;
	pflip = skywin.invert = Flip > 0;
	igmst = gmst(faketime);
	skywin.racen = fixangle((igmst * 15) - siteLon);
	skywin.dlcen = siteLat;
	skywin.scale = 90.0;
	initxform(&skywin);

	// Draw coordinate grid markers, if requested

	if (skyShowCoords && !fChildPreview) {
		int i;
		HPEN cpen, open;
		double epsilon, esin, ecos, eqra, eqdec, eqlat, pera, pedec;
		HFONT ofont, cfont;
#define tickWid	0.75

        cfont = CreateFont(min(15, ((br->right - br->left) / 2) / 12), 0, 0, 0, FW_NORMAL,
        			FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
        			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
        			NULL);
        ofont = SelectObject(hDC, cfont);
    	SetBkMode(hDC, TRANSPARENT);
    	SetTextAlign(hDC, TA_CENTER | TA_BASELINE | TA_NOUPDATECP);
        SetTextColor(hDC, RGB(0, 128, 128));

		open = SelectObject(hDC, (cpen = CreatePen(PS_SOLID, 0, RGB(0, 128, 128))));
		for (i = -1; i <= 1; i += 2) {
			// Equinoctual colures
			plotLine(88.0 * i, 0.0, 90.0 * i, 0.0);
			plotLine(88.0 * i, 180.0, 90.0 * i, 180.0);

			// Solstitial colures
			plotLine(88.0 * i, 90.0, 90.0 * i, 90.0);
			plotLine(88.0 * i, 270.0, 90.0 * i, 270.0);
		}
		
		// Celestial equator
		for (i = 0; i < 360; i += 15) {		// Draw the tick marks at hour angle intervals
			char hlab[4];
			int vx, vy, vis;

            plotLine(0.0, (double) i, 0.0, i + 15.0);
			plotLine(-tickWid, (double) i, tickWid, (double) i);
			sprintf(hlab, Format(42), i / 15);
			xform(tickWid * (pflip ? 1 : -1), (double) i, &vx, &vy, &vis);
			if (vis) {
				TextOut(hDC, vx, vy, hlab, strlen(hlab));
			}

		}
		SelectObject(hDC, open);
		DeleteObject(cpen);
		ticker();

		// Ecliptic

		open = SelectObject(hDC, (cpen = CreatePen(PS_SOLID, 0, RGB(128, 0, 0))));

		epsilon = dtr(obliqeq(faketime));	// Get current obliquity of ecliptic
		esin = sin(epsilon);
		ecos = cos(epsilon);

        SetTextColor(hDC, RGB(128, 0, 0));
		pera = pedec = 0.0;		// Dirty trick: ecliptic intersects equator at 0 longitude !
		for (i = 1; i <= 32; i++) {			// Draw the ecliptic itself
			eqlat = ((PI * 2) / 32.0) * i;
			eqra = fixangle(rtd(atan2(ecos * sin(eqlat), cos(eqlat))));
			eqdec = rtd(asin(esin * sin(eqlat)));
			plotLine(pedec, pera, eqdec, eqra);
			pera = eqra;
			pedec = eqdec;
		}

		for (i = 0; i < 360; i += 15) {		// Draw the tick marks at 15 degree intervals
			char hlab[6];
			int vx, vy, vis;

			eqlat = ((PI * 2) / 360.0) * i;
			pera = fixangle(rtd(atan2((ecos * sin(eqlat) -
					     	(tan(dtr(-tickWid)) * esin)), cos(eqlat))));
			pedec = rtd(asin((esin * sin(eqlat) * cos(dtr(-tickWid))) +
			     		(sin(dtr(-tickWid)) * ecos)));
			eqra = fixangle(rtd(atan2((ecos * sin(eqlat) -
					     	(tan(dtr(tickWid)) * esin)), cos(eqlat))));
			eqdec = rtd(asin((esin * sin(eqlat) * cos(dtr(tickWid))) +
			     		(sin(dtr(tickWid)) * ecos)));
			plotLine(pedec, pera, eqdec, eqra);
			sprintf(hlab, Format(43), i);
			xform(!pflip ? pedec : eqdec, !pflip ? pera : eqra, &vx, &vy, &vis);
			if (vis) {
				TextOut(hDC, vx, vy, hlab, strlen(hlab));
			}

		}
#undef tickWid

		SelectObject(hDC, open);
		DeleteObject(cpen);
		SelectObject(hDC, ofont);
		DeleteObject(cfont);
	}
	ticker();
	if (bailout()) {
		return;
	}

	// Draw constellation names, if requested

	if (skyShowConnames && !fChildPreview) {
		int cntick = 0;		
		char *vfmt = Format(40);

		if (openRSC("Cnames", "DATABASE")) {
			char cl[80], cn[80];
			unsigned int tra;
			int tdec;
			HFONT ofont, cfont;

			if (!skyAlignConnames) {
	            cfont = CreateFont(min(18, ((br->right - br->left) / 2) / 12), 0, 0, 0, FW_NORMAL,
	            			FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
	            			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
	            			NULL);
	            ofont = SelectObject(hDC, cfont);
            }
        	SetBkMode(hDC, TRANSPARENT);
           	SetTextColor(hDC, RGB(255, 255, 0));
        	SetTextAlign(hDC, TA_CENTER | TA_BASELINE | TA_NOUPDATECP);
			while (fgetsRSC(cl, sizeof cl) != NULL) {
				cl[strlen(cl) - 1] = EOS;

				if (strlen(cl) == 0 || isspace(cl[0]) || cl[0] == ';') {
					continue;
				}
				sscanf(cl, vfmt, &tra, &tdec, cn);
				if (tra == 0 && tdec == 0) {
					break;
				}
				if (strchr(cn, '_') != NULL) {
					*strchr(cn, '_') = ' ';
				}
	    		ra = tra / (24000.0 / 360.0);
	    		dec = tdec / (9000.0 / 90.0);
	    		Prd(ra, dec);
				xform(dec, ra, &vx, &vy, &vis);
				if (vis) {
					if (skyAlignConnames) {
						int iangle;
						double eangle;
	
						vx2 = vx - (br->left + (br->right - br->left) / 2);
						vy2 = vy - (br->top + (br->bottom - br->top) / 2);
						if (vx2 == vy2) {
							eangle = 0;
						} else {
							eangle = rtd(atan2((double) vy2, (double) vx2));
							if (eangle >= 0) {
								eangle -= 90;
							} else {
								eangle += 90;
							}
						}
						eangle = - eangle;
						iangle = (int) (eangle * 10); 
			            cfont = CreateFont(min(18, ((br->right - br->left) / 2) / 12), 0, iangle, iangle, FW_NORMAL,
			            			FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
			            			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
			            			NULL);
			            ofont = SelectObject(hDC, cfont);
		            }
					TextOut(hDC, vx, vy, cn, strlen(cn));
					if (skyAlignConnames) {
						SelectObject(hDC, ofont);
						DeleteObject(cfont);
					}
					if ((++cntick % 2) == 0) {
						ticker();
						if (bailout()) {
							break;
						}
					}
				}
			}
			closeRSC();
			if (!skyAlignConnames) {			
				SelectObject(hDC, ofont);
				DeleteObject(cfont);
			}
		}
	}
	ticker();
	if (bailout()) {
		return;
	}

	// Draw constellation outlines, if requested

	if (skyShowConstellations) {
		WORD conoTick = 0;		
		char *vfmt = Format(36);

		if (openRSC("Conlines", "DATABASE")) {
			char cl[80], cn[10];
			unsigned int fra, tra;
			int fdec, tdec;
			double ra2, dec2, cra1, cdec1, cra2, cdec2;
			HPEN cpen, open;

			open = SelectObject(hDC, (cpen = CreatePen(PS_SOLID, 0, RGB(128, 128, 128))));
			while (fgetsRSC(cl, sizeof cl) != NULL) {
				cl[strlen(cl) - 1] = EOS;

				if (strlen(cl) == 0 || isspace(cl[0]) || cl[0] == ';') {
					continue;
				}
				sscanf(cl, vfmt, cn, &fra, &fdec, &tra, &tdec);
				if (fra == 0 && tra == 0) {
					break;
				}
                if ((++conoTick % 40) == 0) {
                	conoTick = 0;
                	ticker();
                }				
	    		ra = fra / (24000.0 / 360.0);
	    		dec = fdec / (9000.0 / 90.0);
	    		Prd(ra, dec);
	    		ra2 = tra / (24000.0 / 360.0);
	    		dec2 = tdec / (9000.0 / 90.0);
	    		Prd(ra2, dec2);
				if (clipr_xform(dec, ra, dec2, ra2, &vx, &vy, &vx2, &vy2, FALSE,
					&cdec1, &cra1, &cdec2, &cra2)) {
					MoveTo(hDC, vx, vy);
					LineTo(hDC, vx2, vy2);
				}
			}
			SelectObject(hDC, open);
			DeleteObject(cpen);
			closeRSC();
		}
	}
	ticker();
	if (bailout()) {
		return;
	}

	// Draw constellation boundaries, if requested

	if (skyShowConbounds && !fChildPreview) {
		WORD conbTick = 0;		
		char *vfmt = Format(39);
		
		if (openRSC("Cbounds", "DATABASE")) {
			char cl[80];
			unsigned int fra, tra;
			int mdraw, fdec, tdec;
			double ra2, dec2, cra1, cdec1, cra2, cdec2;
			HPEN cpen, open;

			open = SelectObject(hDC, (cpen = CreatePen(PS_SOLID, 0, RGB(0, 128, 0))));
			while (fgetsRSC(cl, sizeof cl) != NULL) {
				cl[strlen(cl) - 1] = EOS;

				if (strlen(cl) == 0 || isspace(cl[0]) || cl[0] == ';') {
					continue;
				}
                if ((++conbTick % 20) == 0) {
					conbTick = 0;
                	ticker();
                	if (bailout()) {
                		break;
                	}
                }				
				sscanf(cl, vfmt, &mdraw, &tra, &tdec);
				if (mdraw == 2) {
					break;
				} else if (mdraw) {
		    		ra = fra / (24000.0 / 360.0);
		    		dec = fdec / (9000.0 / 90.0);
		    		Prd(ra, dec);

		    		ra2 = tra / (24000.0 / 360.0);
		    		dec2 = tdec / (9000.0 / 90.0);
		    		Prd(ra2, dec2);
					if (clipr_xform(dec, ra, dec2, ra2, &vx, &vy, &vx2, &vy2, FALSE,
						&cdec1, &cra1, &cdec2, &cra2)) {
						drawcurveline(cdec1, cra1, cdec2, cra2, vx, vy, vx2, vy2, 0, FALSE, 0, hDC);
					}
				}
				fra = tra;
				fdec = tdec;
			}
			SelectObject(hDC, open);
			DeleteObject(cpen);
			closeRSC();
		}
	}
	ticker();
	if (bailout()) {
		return;
	}

	// Scan the star map and draw the stars

	smap = LockResource(starMapRes);
	if (smap != NULL) {
		struct starMapData sd;
		struct starMapData *smp = (smap == NULL) ? &sd : smap;
		int imag;
		short ipmra, ipmdec;
		unsigned int ilimag;
		HFONT ofont, cfont, gfont;
		int starTick = 0;

        cfont = CreateFont(min(15, ((br->right - br->left) / 2) / 12), 0, 0, 0, FW_NORMAL,
        			FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
        			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
        			NULL);
        gfont = CreateFont(min(15, ((br->right - br->left) / 2) / 12), 0, 0, 0, FW_NORMAL,
        			FALSE, FALSE, FALSE, SYMBOL_CHARSET, OUT_DEFAULT_PRECIS,
        			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
        			rstring(IDS_GREEK_FONTNAME));
        ofont = SelectObject(hDC, cfont);
    	SetBkMode(hDC, TRANSPARENT);
    	SetTextAlign(hDC, TA_LEFT | TA_BASELINE | TA_NOUPDATECP);
       	SetTextColor(hDC, PALETTERGB(255, 128, 128));

		if (fChildPreview) {
			limag =  2.0;
		}
        ilimag = (unsigned int) ((limag + 1.5) * 100); // For quick tests against limiting magnitude
		while (TRUE) {
			char *sname;

          	memset(smex, 0, sizeof smex);
          	sname = ((char *) smp) + sizeof(struct starMapData);
          	if (smp->mag & 0x2000) {
          		memcpy(smex, sname, 3);
          		sname += 3;
          	}
          	if (smp->mag & 0x1000) {
          		memcpy(&ipmra, sname, sizeof ipmra);
          		memcpy(&ipmdec, sname + sizeof ipmra, sizeof ipmdec);
          		sname += sizeof ipmra + sizeof ipmdec;
          	} else {
          		ipmra = ipmdec = 0;
          	}
    		ra = smp->lon / (65536.0 / 360.0) +
    			(((faketime - PrecEpoch) / (JulianCentury / 100)) * (ipmra / (1000.0 * 60 * 60)));
			dec = smp->lat / (32767.0 / 90.0) +
				(((faketime - PrecEpoch) / (JulianCentury / 100)) * (ipmdec / (1000.0 * 60 * 60)));
    		Prd(ra, dec);

			xform(dec, ra, &vx, &vy, &vis);
			if (vis) {
				double mbase = 3.5;
                
                if ((++starTick % 40) == 0) {
                	starTick = 0;
                	ticker();
                	if (bailout()) {
                		break;
                	}
                } 
	    		mag = ((smp->mag & 0xFFF) / 100.0) - 1.5;
				if (!fChildPreview) {
					int named = FALSE;

					/* If star is bright enough to meet the "show name" criterion
					   and it has a name, display it. */

					if ((smp->mag & 0x8000) && skyShowName && (mag < skyNameMag)) {
						SelectObject(hDC, cfont);
	       		     	TextOut(hDC, vx + 8, vy + 2, sname, lstrlen(sname));
	       		     	named = TRUE;
					}

					// See if the Bayer/Flamsteed code should be drawn

					if ((smex[0] != 0) && skyShowBflam && (mag < skyBflamMag)) {
						SelectObject(hDC, smex[0] < 32 ? gfont : cfont);
						if (smex[0] < 32) {
							smex[0] += ('a' - 1);
						}
						TextOut(hDC, vx + (named ? -16 : 8), vy + 2,
							smex, strlen(smex));

					}
				}

				// Finally, paint the star according to its magnitude and the quality setting

				imag = ((int) ELEMENTS(starIcons)) - ((int) (mag - BrightestMag));
				if (limag < (ELEMENTS(starIcons) - 2)) {
					imag -= (ELEMENTS(starIcons) - ((int) limag)) - 1;
				}
				imag = min(max(imag, 0), ELEMENTS(starIcons) - 1);
				DrawIcon(hDC, vx - 16, vy - 16, starIcons[imag]);
			}

			/* Quit if we hit the end of file or exceed the limiting magnitude.
			   Note the assumption that the catalogue is sorted by magnitude! */

    		if ((smp->mag & 0x4000) || ((smp->mag & 0xFFF) > ((int) ilimag))) {
    			break;
    		}
    		if (smap != NULL) {
    			smp = (struct starMapData *) (sname + ((smp->mag & 0x8000) ? (strlen(sname) + 1) : 0));
    		}
		}

		SelectObject(hDC, ofont);
		DeleteObject(cfont);
		DeleteObject(gfont);

		UnlockResource(starMapRes);
	}
	ticker();
	if (bailout()) {
		return;
	}

	// Plot deep sky objects, if requested

	if (skyShowDeep && !fChildPreview) {
		char *dcode = rstring(IDS_DEEP_CODES), *vfmt = Format(41);
		HFONT ofont, cfont;

        cfont = CreateFont(min(15, ((br->right - br->left) / 2) / 12), 0, 0, 0, FW_NORMAL,
        			FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
        			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
        			NULL);
        ofont = SelectObject(hDC, cfont);
    	SetBkMode(hDC, TRANSPARENT);
       	SetTextColor(hDC, PALETTERGB(255, 128, 128));
    	SetTextAlign(hDC, TA_LEFT | TA_BASELINE | TA_NOUPDATECP);

		if (openRSC("Deepobj", "DATABASE")) {
			char cl[132], cn[80], dc;
			double ra, dec, mag;

			while (fgetsRSC(cl, sizeof cl) != NULL) {
				cl[strlen(cl) - 1] = EOS;

				if (strlen(cl) == 0 || isspace(cl[0]) || cl[0] == ';') {
					continue;
				}
				if (sscanf(cl, vfmt, &ra, &dc, &dec, &dc, &mag, &dc, cn) < 7) {
					continue;
				}
				if (mag == -99 || mag > skyDeepMag) {
					break;
				}
	    		Prd(ra, dec);
				xform(dec, ra, &vx, &vy, &vis);
				if (vis) {
					char *objclass, *popname;
					int i;
					HICON di = NULL;

					popname = strchr(cn, ',');
					if (popname != NULL) {
						*popname++ = EOS;
						if ((objclass = strchr(popname, ',')) != NULL) {
							*strchr(popname, ',') = EOS;
						}
						if (objclass != NULL) {
							objclass++;
							if (strchr(objclass, ',') != NULL) {
								*strchr(objclass, ',') = EOS;
							}
							for (i = 0; i < ELEMENTS(deepIcons); i++) {
								if (strncmp(objclass, dcode + (i * 2), 2) == 0) {
									di = deepIcons[i];
									break;
								}
							}
							if (di != NULL) {
								char *dname = strlen(popname) > 0 ? popname : cn;

								DrawIcon(hDC, vx - 16, vy - 16, di);
	       		     			TextOut(hDC, vx + 8, vy + 2, dname, lstrlen(dname));
							}
						}
					}

				}
			}
			SelectObject(hDC, ofont);
			DeleteObject(cfont);
			closeRSC();
		}
	}
	ticker();
	if (bailout()) {
		return;
	}

	/*  Now paint the Sun, Moon, and planets on the map.  */

    if (skyShowPlanets && !fChildPreview) {
		for (i = 0; i <= 9; i++) {
			if (((planet_info[i].alt > 0.0) || !skyTel) && ((i < 9) || (planet_info[i].hrv > 0))) {
	    		ra = dtr(planet_info[i].ra);
	    		dec = dtr(planet_info[i].dec);

	    		/* Note that since planetary positions are computed for the equinox
	    		   of the date, they are *not* precessed to the current equinox. */

				xform(rtd(dec), rtd(ra), &vx, &vy, &vis);
				if (!vis) {
					continue;
				}

				if (i == 3 && micon != NULL) {
					// Use the moon icon to show current phase
					DrawIcon(hDC, vx - 16, vy - 13, micon);
				} else {
					DrawIcon(hDC, vx - 16, vy - 16, PlanetIcon(i));
				}
			}
		}
	}
#undef Prd
}

/*  RELEASESKYBITMAPS  --  Release backing bitmaps for sky map.  */

static void releaseSkyBitmaps(void)
{
	if (smbitmap != NULL) {
		DeleteObject(smbitmap);
		smbitmap = NULL;
	}
}

/*  SKYWNDPROC  --  Window procedure for sky map.  */

WindowProc(ScreenSaverProc)
{
    HDC hDC, hMemDC;
    PAINTSTRUCT ps;         			// holds PAINT information
    static short cxClient, cyClient;    // Window size
    HBITMAP obmap;
    HBRUSH obrush;
    HPEN open, bpen;
    HPALETTE opal, mopal;
    HRGN brgn;
    RECT cr;							// Client rectangle
    int minsize;						// Short direction in window
    static int offX = 0, offY = 0;		// Offset of image within window
    static RECT br;						// Sub-rectangle containing map (save for clicking)
    static long cctime;					// Time to display (real or simulated) */
    static long rptime = 0;				// Last time image shifted
    static DWORD calctime;				// Calculation time in ticks
    static WORD minlapse = 0;			// Minutes elapsed since last calculation
	TEXTMETRIC tm;
	int charoff, brx, bry;
    struct tm ct, lt;
    double jt;
    BOOL repositioned = FALSE;
    static BOOL busy = FALSE;
    static BOOL noMem = FALSE;			// Unrecoverable out of memory situation
	static char stdname[128], dstname[128];	// Standard and Daylight time zone names

    switch (Message) {

        case WM_CREATE:
#ifdef Mdiag        
			OutputDebugString("\r\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
#endif
			/*	Reinitialise static variables so we're restartable when
				tested under Scream Saver.  */
			offX = offY = 0;
			rptime = 0;
			minlapse = 0;
			busy = FALSE;
			noMem = FALSE;
			firstTick = nextTick = FALSE;
			calculationNumber = 0;
			bailedOut = FALSE;
			smbitmap = NULL;
			starPal = NULL;
			
            hDC = GetDC(hWnd);
			SetCursor(NULL);			// Zap cursor before first draw
            GetIniEntries();
            GetIniSettings();			// Get setup defaults
            siteLat = latDeg + (latMin / 60.0) + (latSec / (60.0 * 60.0));
            siteLon = lonDeg + (lonMin / 60.0) + (lonSec / (60.0 * 60.0));
			GetTimeZoneInformation(&tzInfo);
			wcstombs(stdname, tzInfo.StandardName, sizeof stdname);
			tzName[FALSE] = stdname;
			wcstombs(dstname, tzInfo.DaylightName, sizeof dstname);
			tzName[TRUE] = dstname;
            localtimeformat();
            srand((unsigned) (GetTickCount() & 0xFFFF));
            starMapRes = LoadResource(hInst, starMapHandle = FindResource(hInst, starCatResName[starCatalogue], "STARMAP"));
            loadPlanetIcons();
            ReleaseDC(hWnd, hDC);
            SetTimer(hWnd, 1, 1000u, NULL);
            break;

        case WM_COMPACTING:
        	releaseSkyBitmaps();
        	break;

        case WM_DESTROY:
            KillTimer(hWnd, 1);
        	releaseSkyBitmaps();
        	FreeResource(starMapRes);
        	unloadPlanetIcons();
            if (micon != NULL) {
                DestroyIcon(micon);
                micon = NULL;
            }
        	if (starPal != NULL) {
        		DeleteObject(starPal);
        		starPal = NULL;
        	}
			break;
       
        case WM_TIMER:
			if (!busy) {
				busy = TRUE;
				time(&cctime);
	            if (firstTick) {
		            hDC = GetDC(hWnd);
		        	GetClientRect(hWnd, &cr);
		        	minsize = min(cr.right, cr.bottom);
		        	cr.left = offX;
		        	cr.top = offY;
			        cr.right = cr.left + minsize;
			        cr.bottom = cr.top + minsize;
					if (!fChildPreview) {					      
						set_tm_time(&lt, TRUE);
						set_tm_time(&ct, FALSE);
		       			ssavetime(hDC, lt, ct, cr, skycolour);	// Update local and universal time
					}
		            ReleaseDC(hWnd, hDC);
		        	if ((cctime % 60 == 0) || nextTick) {
		        		minlapse++;
		        		
		        		/* The following code is a little odd.  What we're doing it
		        		   guaranteeing that on slow machines, the duty cycle of
		        		   calculation will not exceed 50%.  We do this by deferring
		        		   the once-per-minute update, if necessary, until the time
		        		   lapse since the last update is at least twice as long as
		        		   we spent on the last calculation. */
		        		   
		        		if (((minlapse * 60000uL) >= (calctime / 2)) || nextTick) {
		        			nextTick = FALSE;
		        			minlapse = 0;
							InvalidateRect(hWnd, NULL, FALSE);
						}
					}
				} else {
					firstTick = TRUE;
					nextTick = TRUE;
					InvalidateRect(hWnd, NULL, FALSE);
				}
				busy = FALSE;
			}
			
			/*  If we've run out of memory (whether during initialisation or
				during generation),  display a randomly-positioned out of
				memory message on the screen.  */
				 
			if (noMem) {
				char s[132];
				int tx, ty;
				
				if (cctime % 10 == 0) {
					RECT rc;
					SIZE sz;

			        hDC = GetDC(hWnd);
		    		GetClientRect(hWnd, &rc);
	  				LoadString(hMainInstance, idsInsuffMemory, s, sizeof s);
	  				GetTextExtentPoint32(hDC, s, lstrlen(s), &sz);
	  				SetBkMode(hDC, TRANSPARENT);
	  				tx = (int) (((rc.right - sz.cx) * ((long) rand())) / RAND_MAX); 
	  				ty = sz.cy + (int) (((rc.bottom - (2 * sz.cy)) * ((long) rand())) / RAND_MAX);
	  				rc.top = sz.cy;
		    		FillRect(hDC, &rc, (HBRUSH) GetStockObject(BLACK_BRUSH));
	  				skycolour = RGB(255, 255, 255);  
	  				SetTextColor(hDC, skycolour);
	  				SetTextAlign(hDC, TA_NOUPDATECP);
	  				TextOut(hDC, tx, ty, s, lstrlen(s));
		        	ReleaseDC(hWnd, hDC);
		        }
			}			
        	return 0;

        case WM_PAINT:        
            memset(&ps, 0, sizeof(PAINTSTRUCT));
            hDC = BeginPaint(hWnd, &ps);
            if (!firstTick || noMem) {
	            ValidateRgn(hWnd, NULL);
	            EndPaint(hWnd, &ps);
	            break;
            }        
			set_tm_time(&lt, TRUE);
			set_tm_time(&ct, FALSE);
			time(&cctime);
			jt = faketime = jtime(&ct);
			calcPlanet(jt);
            updMoonIcon(hDC, jt);
            GetClientRect(hWnd, &cr);
            minsize = min(cr.right, cr.bottom);
            
            //	Save time and position to update ticker during calculation
            
            tickerLast = cctime;
        	tickerRect.left = offX;
        	tickerRect.top = offY;
	        tickerRect.right = tickerRect.left + minsize;
	        tickerRect.bottom = tickerRect.top + minsize;					      
            
            /* If it's time, randomly reposition the circular sky map within the
               rectangular window.  This prevents burning in the phosphor by
               the constant parts of the map. */ 

            if ((cctime - rptime) > (RepositionTime * 60L)) {
            	offX = (int) (((cr.right - minsize) * ((long) rand())) / RAND_MAX);
            	offY = (int) (((cr.bottom - minsize) * ((long) rand())) / RAND_MAX);
            	
            	/* Sigh.  If this is the first time, we must set the ticker
            	   location to the window where we're going to paint the initial
            	   sky map. */
            	   
            	if (rptime == 0) {
		        	tickerRect.left = offX;
		        	tickerRect.top = offY;
			        tickerRect.right = tickerRect.left + minsize;
			        tickerRect.bottom = tickerRect.top + minsize;					      
		        }
            	rptime = cctime;
            	repositioned = TRUE;
            }
            cr.right = cr.bottom = minsize;		// Now restrict to actual draw window

			if (smbitmap == NULL) {
				smbitmap = CreateCompatibleBitmap(hDC, minsize, minsize);
				if (smbitmap == NULL) {
					// Insufficient memory to allocate bitmap.  We're stuck
					noMem = TRUE;
				}
			}
#ifdef Mdiag			
			if (smbitmap == NULL) OutputDebugString("smbitmap alloc failed.\r\n");
#endif        
	    	
	    	if (smbitmap != NULL) {
            
	            /* Build colour palette to define the colours we'll use in plotting
	               the star map. */
	
				if (starPal == NULL) {
		    		char palbuf[sizeof(LOGPALETTE) + 8 * sizeof(PALETTEENTRY)];
		    		LOGPALETTE *lp = (LOGPALETTE *) palbuf;
	
#define sC(n, r, g, b)		lp->palPalEntry[n].peRed = r; \
							lp->palPalEntry[n].peGreen = g; \
							lp->palPalEntry[n].peBlue = b; \
							lp->palPalEntry[n].peFlags = PC_NOCOLLAPSE
								
	    			lp->palNumEntries = 8;
	    			lp->palVersion = 0x300;
	    			sC(0,   0,   0,   0);
	    			sC(1, 128, 128, 128);
	    			sC(2, 255, 255,   0);
	    			sC(3, 255, 128, 128);
	    			sC(4,   0, 128, 128);
	    			sC(5, 128,   0,   0);
	    			sC(6,   0, 128,   0);
	    			sC(7,   0,   0, 255);
	    			sC(8,   0,   0, 128);
#undef sC    			
	    			
	    			starPal = CreatePalette(lp);
#ifdef Mdiag
					if (starPal == NULL) {
						OutputDebugString("CreatePalette(lp) failed.\r\n");
					}
#endif
		    	}
		    	if (starPal != NULL) {
		    		opal = SelectPalette(hDC, starPal, FALSE);
		    		RealizePalette(hDC);
		    	}
		    	        
				hMemDC = CreateCompatibleDC(hDC);
#ifdef Mdiag			
				if (hMemDC == NULL) OutputDebugString("hMemDC create failed.\r\n");
#endif			
				if (starPal != NULL) {
		    		mopal = SelectPalette(hMemDC, starPal, FALSE);
		    	}
				obmap = SelectObject(hMemDC, smbitmap);        
	
	            calctime = GetTickCount();
	            Flip = (siteLat >= 0) ? 1 : -1;
	        	GetTextMetrics(hDC, &tm);
	        	FillRect(hMemDC, &cr, GetStockObject(BLACK_BRUSH));
	        	br = cr;
	
	        	/* Label cardinal points on the map with text and shrink the logical
	        	   map so it doesn't overlap the labels. */
	
	        	charoff = (5 * max(tm.tmHeight + 1, tm.tmMaxCharWidth + 1)) / 4;
	        	br.left += charoff;
	        	br.right -= charoff;
	        	br.top += charoff;
	        	br.bottom -= charoff;
	
	            // Determine legend colour based on altitude of Sun
	
	            skycolour = (planet_info[0].alt < -18.0) ?
	            											RGB(128, 128, 128) :           // Night
	            					      ((planet_info[0].alt > -1) ? RGB(0, 0, 255) :    // Day
	            					        ((planet_info[0].alt > -6) ? RGB(128, 0, 0) :  // Dawn/Sunset
	            					      		RGB(0, 0, 128)                             // Twilight
	            					        )
	            					      );
	
	        	SetBkMode(hMemDC, TRANSPARENT);
	        	SetTextColor(hMemDC, skycolour);
	        	SetTextAlign(hMemDC, TA_CENTER | TA_BOTTOM | TA_NOUPDATECP);
	        	TextOut(hMemDC, (br.left + br.right) / 2, br.top - 1,
	        		rstring(Flip > 0 ? IDS_NORTH : IDS_SOUTH), 1);
	        	SetTextAlign(hMemDC, TA_CENTER | TA_TOP | TA_NOUPDATECP);
	        	TextOut(hMemDC, (br.left + br.right) / 2, br.bottom + 1,
	        		rstring(Flip > 0 ? IDS_SOUTH : IDS_NORTH), 1);
	        	SetTextAlign(hMemDC, TA_CENTER | TA_BASELINE | TA_NOUPDATECP);
	        	TextOut(hMemDC, (cr.left + br.left) / 2,
	        					((br.top + br.bottom) + tm.tmHeight) / 2,
	        		rstring(Flip > 0 ? IDS_EAST : IDS_WEST), 1);
	        	SetTextAlign(hMemDC, TA_CENTER | TA_BASELINE | TA_NOUPDATECP);
	        	TextOut(hMemDC, (cr. right + br.right) / 2,
	        					((br.top + br.bottom) + tm.tmHeight) / 2,
	        		rstring(Flip > 0 ? IDS_WEST : IDS_EAST), 1);
	
	        	// Draw "compass rose" marks for NW, SW, NE, and SE
	
	            bpen = CreatePen(PS_SOLID, 0, skycolour);
	            brx = (br.right - br.left) / 2;
	            bry = (br.bottom - br.top) / 2;
	            open = SelectObject(hMemDC, bpen);
	            MoveTo(hMemDC, br.left + (1 * brx) / 4,
	            			   br.top + (1 * bry) / 4);
	            LineTo(hMemDC, br.left + (7 * brx) / 4,
	            			   br.top + (7 * bry) / 4);
	            MoveTo(hMemDC, br.left + (1 * brx) / 4,
	            			   br.top + (7 * bry) / 4);
	            LineTo(hMemDC, br.left + (7 * brx) / 4,
	            			   br.top + (1 * bry) / 4);
	
	            obrush = SelectObject(hMemDC, GetStockObject(BLACK_BRUSH));
	        
	            Ellipse(hMemDC, br.left, br.top, br.right, br.bottom);
	            SelectObject(hMemDC, obrush);
	            SelectObject(hMemDC, open);
	            DeleteObject(bpen);
	       		
	            obrush = SelectObject(hMemDC, GetStockObject(WHITE_BRUSH));
	            open = SelectObject(hMemDC, GetStockObject(WHITE_PEN));
	            
	            /* To obtain precise clipping of objects on the star map at
	               the horizon, we create an elliptical (actually circular)
	               clipping region.  For large screens, this can take up a
	               lot of memory and, when memory is tight, may very well
	               fail.  That isn't the end of the world, however--all it
	               means is that legends may extend slightly outside the
	               circle of the map, so we soldier on without the clipping
	               region if we can't obtain it.  It is also possible that
	               if we successfully obtain the region, the attempt to select
	               it into the device context (which makes a copy of the
	               region) may fail.  Amazingly, or perhaps not to experienced
	               Windows hands, this corrupts the clipping region of the
	               HDC and causes nothing to be drawn.  We have to detect
	               this case and explicitly select a NULL clipping region to
	               avoid having a blank map drawn. */
	        
	        	brgn = CreateEllipticRgn(br.left, br.top, br.right, br.bottom);        
	        	if (brgn != NULL) {
	        		int status = SelectClipRgn(hMemDC, brgn);
	
	        		DeleteObject(brgn);		// Select makes a copy, so we can delete now
					if (status == ERROR) {        		
#ifdef Mdiag
	                	OutputDebugString("BRGN select failed.\r\n");
#endif
						status = SelectClipRgn(hMemDC, NULL);
#ifdef Mdiag
						if (status == ERROR) {        		
		 	               	OutputDebugString("Recovery: NULL region reselect failed.\r\n");
		 	            }
#endif
					}        		
	        	}
#ifdef Mdiag        	
				else OutputDebugString("BRGN alloc failed.\r\n");
#endif
	            hTickerDC = hDC;
	            paintSky(TRUE, hMemDC, &br, skyLimitMag);
	        	SelectClipRgn(hMemDC, NULL);
	            SelectObject(hMemDC, obrush);
	            SelectObject(hMemDC, open);
	            
		        calctime = GetTickCount() - calctime;
	
				if (!bailedOut) {				// Skip paint if we're bailing out
	            
		            /* Paint time into the bitmap so the time doesn't disappear
		               momentarily when we copy in the updated bitmap. */
			        
					if (!fChildPreview) {    
						set_tm_time(&lt, TRUE);
						set_tm_time(&ct, FALSE);
		    			SetTextAlign(hMemDC, TA_LEFT | TA_TOP | TA_NOUPDATECP);
		       			ssavetime(hMemDC, lt, ct, cr, skycolour);	// Display local and universal time
							
						//	If requested, show calculation number and time in lower left corner of map.
							
						if (skyShowTiming) {
		            		char s[40];
			            	
			            	calculationNumber++;	
		            		wsprintf(s, Format(5), calculationNumber, calctime / 1000,
		            			calctime % 1000);
							SetTextColor(hMemDC, skycolour);
							DrawText(hMemDC, s, strlen(s), &cr, DT_LEFT | DT_BOTTOM | DT_NOCLIP | DT_SINGLELINE);
						}
					}
		
		            GetClientRect(hWnd, &cr);		// Recover actual full window rect
		
		            if (repositioned) {
		            	BitBlt(hDC, 0, 0, cr.right, cr.bottom, 0, 0, 0, BLACKNESS);
		            }
		            BitBlt(hDC, offX, offY, minsize, minsize, hMemDC, 0, 0, SRCCOPY);
	            }
#ifdef Mdiag            
				else OutputDebugString("Bailed out.\r\n");
#endif			
	
				/* Reset bailedOut here just in case we see a bogus event that
				   triggers our bailout but doesn't actually terminate the
				   screen saver main loop.  This allows us to recover as of the
				   next update. */
				   
				bailedOut = FALSE;            
	            if (starPal != NULL) {
	            	SelectPalette(hMemDC, mopal, FALSE);
	            	SelectPalette(hDC, opal, FALSE);
	            }
	            SelectObject(hMemDC, obmap);
	            DeleteDC(hMemDC);
            }
            ValidateRgn(hWnd, NULL);

            /* Inform Windows painting is complete */
            EndPaint(hWnd, &ps);
            return 0;
    }
	return DefScreenSaverProc(hWnd, Message, wParam, lParam);
}

